from interface import RoseDriverImpl


class RoseDriver(RoseDriverImpl):

    def __init__(self, url, port, auth):
        super(RoseDriver, self).__init__(url, port, auth)

    def get(self, database, collection, identifier) -> dict:
        """
        Sends a "get" request to the websocket and retrieves the requested data

        :param database: database holding the data.
        :type database: str
        :param collection: collection holding the data
        :type collection: str
        :param identifier: name or identifier of the data
        :type identifier: str
        :return: dict converted from JSON
        :rtype: dict
        """
        data = self.loop.run_until_complete(self._get_impl(database, collection, identifier))
        return data

    def add(self, database, collection, identifier, JSON) -> dict:
        """
        Adds an item to the database in the form of an JSON object
        :param database: the database to place the data.
        :type database: str
        :param collection: collection to place the data
        :type collection: str
        :param identifier: name or identifier of the file
        :type identifier: str
        :param JSON: JSON/Dict object to be stored
        :type JSON: dict, JSON
        :return: response [dict]
        :rtype: dict
        """
        data = self.loop.run_until_complete(self._add_impl(database, collection, identifier, JSON))
        return data

    def remove(self, database, collection, identifier, *key) -> dict:
        """
        Removes *keys(value) from the given Item
        :param database: the database of the file containing the key/value to remove.
        :type database: str
        :param collection: collection of the file containing the key/value to remove.
        :type collection: str
        :param identifier: name or identifier of the file containing the key/value to remove.
        :type identifier: str
        :param key: key to remove from the identifier/file
        :type key: any
        :return: dict
        :rtype: dict
        """
        data = self.loop.run_until_complete(self._remove_impl(database, collection, identifier, key))
        return data

    def remove_item(self, database, collection, identifier) -> dict:
        """
        Removes an Identifier/file from a given collection.
        :param database: the database of the file
        :type database: str
        :param collection: collection of the file
        :type collection: str
        :param identifier: name or identifier of the file
        :type identifier: str
        :return: dict
        :rtype: dict
        """
        data = self.loop.run_until_complete(self._remove_item_impl(database, collection, identifier))
        return data

    def remove_Collection(self, database, collection) -> dict:
        data = self.loop.run_until_complete(self._remove_Collection_impl(database, collection))
        return data

    def remove_Database(self, database) -> dict:
        data = self.loop.run_until_complete(self._remove_Database_impl(database))
        return data



import asyncio
import random
import socket
import string
import json
from websockets import connect


class RoseDriverImpl:

    def __init__(self, url, port, auth):
        self.__url = f"ws://{url}:{port}"
        self.__auth = f"{auth}"
        self.ws = None
        self.loop = asyncio.get_event_loop()
        self.loop.run_until_complete(self.build())
        self.size = 24
        self.delay = 0.2

    async def build(self):
        try:
            print(f"Connecting to [{self.__url}]")
            self.ws = await connect(self.__url)
            print("Connected")
        except socket.gaierror:
            print("[Error]: Failed to parse the address. Perhaps there is a typo in the Websocket address?")
        except (ConnectionRefusedError, ConnectionError) as e:
            print(f"[Error]: {e} ")

    async def _get_impl(self, database, collection, identifier) -> dict:
        json_send = json.dumps({"authorization": self.__auth, "method": "get", "database": database,
                                "collection": collection, "identifier": identifier, "unique": self._id_generator()})
        await self.ws.send(json_send)
        data = await self._listener()
        return data

    async def _add_impl(self, database, collection, identifier, JSON) -> dict:
        json_send = json.dumps({"authorization": self.__auth, "method": "add", "database": database,
                                "collection": collection, "identifier": identifier, "value": JSON,
                                "unique": self._id_generator()})
        await self.ws.send(json_send)
        data = await self._listener()
        return data

    async def _remove_impl(self, database, collection, identifier, *key) -> dict:
        json_send = json.dumps({"authorization": self.__auth, "method": "delete", "database": database,
                                "collection": collection, "identifier": identifier, "key": key,
                                "unique": self._id_generator()})
        await self.ws.send(json_send)
        data = await self._listener()
        return data

    async def _remove_item_impl(self, database, collection, identifier) -> dict:
        json_send = json.dumps({"authorization": self.__auth, "method": "delete", "database": database,
                                "collection": collection, "identifier": identifier,
                                "unique": self._id_generator()})
        await self.ws.send(json_send)
        data = await self._listener()
        return data

    async def _remove_Collection_impl(self, database, collection) -> dict:
        json_send = json.dumps({"authorization": self.__auth, "method": "delete", "database": database,
                                "collection": collection,
                                "unique": self._id_generator()})
        await self.ws.send(json_send)
        data = await self._listener()
        return data

    async def _remove_Database_impl(self, database) -> dict:
        json_send = json.dumps({"authorization": self.__auth, "method": "delete", "database": database,
                                "unique": self._id_generator()})
        await self.ws.send(json_send)
        data = await self._listener()
        return data

    # simple method for closing the connection safely
    async def close(self):
        try:
            return await self.ws.close()
        except Exception as e:
            print(f"[Error]: {e}")

    # small "listener" function
    async def _listener(self) -> dict:
        while True:
            try:
                recv_data = await self.ws.recv()
                if "session" in recv_data:
                    await asyncio.sleep(self.delay)
                    continue
                else:
                    data = json.loads(recv_data)
                    return data["response"]
            except KeyError:
                print("[Error]: A delay in response caused the listener to not receive the response"
                      " before handing it over. It is advised to change the Drivers [delay] attribute higher.")
                continue

    # generates the unique string
    def _id_generator(self, chars=string.ascii_uppercase + string.digits) -> str:
        return ''.join(random.choice(chars) for _ in range(self.size))
